from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR3, namely changes the `type` slot to be NucleotideSequencing or 
    MassSpectrometry for the data_generation_set documents.
    """

    _from_version = "X"
    _to_version = "PR3"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        self.adapter.process_each_document(collection_name="data_generation_set", pipeline=[self.specify_data_gen_type])

    def specify_data_gen_type(self, data_gen_doc: dict) -> dict:
        r"""
        Changes the `type` slot to be either nmdc:NucleotideSequencing or nmdc:MassSpectrometry based 
        on the value of the analyte_category slot of the DataGeneration document

        >>> m = Migrator()  
        >>> m.specify_data_gen_type({'id': 123, 'analyte_category': 'nom', 'type': 'nmdc:DataGeneration'})  
        {'id': 123, 'analyte_category': 'nom', 'type': 'nmdc:MassSpectrometry'}
        >>> m.specify_data_gen_type({'id': 234, 'analyte_category': 'metagenome', 'type': 'nmdc:DataGeneration'})
        {'id': 234, 'analyte_category': 'metagenome', 'type': 'nmdc:NucleotideSequencing'}
        """


        nucleotide_seqs = ["metagenome", "metatranscriptome"]
        mass_spec = ['metaproteome', 'metabolome', 'lipidome', 'nom']

        if data_gen_doc["analyte_category"] in nucleotide_seqs:
            data_gen_doc["type"] = 'nmdc:NucleotideSequencing'
        elif data_gen_doc["analyte_category"] in mass_spec:
            data_gen_doc["type"] = 'nmdc:MassSpectrometry'
        else:
            self.logger.error(
                    f"ERROR: analyte_category for {data_gen_doc['id']} is {data_gen_doc['analyte_category']}, which is not one of {nucleotide_seqs} or {mass_spec}")

        return data_gen_doc



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data between two schemas.

    TUTORIAL: This is an example of a "migrator" class.
              It was designed for use during developer training and
              to serve as a template for production "migrator" classes.
    """

    """
    TUTORIAL: These two strings tell people the versions of the schema this
              "migrator" class was designed to migrate data between.
              
              -->  As part of creating a new "migrator" class, you will
                   populate its `_from_version` and `_to_version` strings.
    """
    _from_version = "1.0.0"  # https://github.com/microbiomedata/nmdc-schema/blob/1.0.0/
    _to_version = "EXAMPLE"  # in practice, this would be a real schema version; e.g., "7.8.0"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        TUTORIAL: This is the `upgrade` function. You can think of it as the "main" function or
                  "entry point" of the migrator. Its job is to transform the database from
                  conforming to the original schema to conforming to the new schema.

                  The `upgrade` function accesses the database via an "adapter", instead of
                  accessing the database directly. That allows the same migrator to be used
                  with different types of data stores (e.g. MongoDB database, Python dictionary).

              --> As part of creating a new "migrator" class, you will implement an `upgrade` function.
        """

        # Process each document in the specified collection.
        #
        # Note: In this example, we will pass each document in the `study_set` collection through
        #       a processing pipeline that consists of a single function: `self.allow_multiple_names`.
        #
        self.adapter.process_each_document("study_set", [self.allow_multiple_names])

        # Create and populate a collection that doesn't exist in the `_from_version` schema.
        self.adapter.create_collection("comment_set")
        self.adapter.insert_document("comment_set", {"id": 1, "text": "Hello"})
        self.adapter.insert_document("comment_set", {"id": 2, "text": "Goodbye"})

        # Advanced: Create and populate another new collection; based upon the documents in a _different_ collection.
        self.adapter.create_collection("report_set")
        self.adapter.process_each_document("comment_set", [self.create_report_based_upon_comment])

    def allow_multiple_names(self, study: dict) -> dict:
        """
        TUTORIAL: This is an example "transformation" function within the class.
                  Its job is to transform a dictionary that conforms to the
                  initial schema (in this case, version "1.0.0"), into one
                  that conforms to the target schema (in this case,
                  version "EXAMPLE"). That might involve adding a field,
                  converting a string into a list of strings, etc.

                  When this "transformation" function runs, it does three things:
                  1. It accepts a single dictionary as a parameter.
                  2. It transforms that dictionary.
                  3. It returns the transformed dictionary.

              --> As part of creating a new "migrator" class, you will
                  typically implement one or more "transformation" functions.

        TUTORIAL: The remaining part of this comment consists of several "doctests".
                  You can read more about doctests in the official Python documentation:
                  https://docs.python.org/3/library/doctest.html

                  Each doctest consists of (a) a snippet of Python code that calls this
                  function, followed on the next line by (b) whatever the author of the
                  test expects this function to return, if anything.

                  People can then read and run those doctests in order to learn about
                  what this function does / how this function behaves.

              --> As part of implementing a "transformation" function, you will
                  typically write a few doctests in the docstring of that function.

        >>> m = Migrator()  # creates a class instance on which we can call this function (i.e. this method)
        >>> m.allow_multiple_names({'id': 123})  # test: creates an empty `names` list
        {'id': 123, 'names': []}
        >>> m.allow_multiple_names({'id': 123, 'name': 'My project'})  # test: transfers existing name to `names` list
        {'id': 123, 'names': ['My project']}
        >>> m.allow_multiple_names({'id': 123, 'name': 'My project', 'foo': 'bar'})  # test: preserves other keys
        {'id': 123, 'foo': 'bar', 'names': ['My project']}
        """

        # optional log message
        self.logger.info(f"Transforming study having id: {study['id']}")

        # Transform the dictionary.
        #
        # TUTORIAL: In this example, I am creating a new field named `names`, whose value is an empty list;
        #           then, if the original study has a `name` value, I'm storing that in the list and then
        #           deleting the `name` field.
        #
        study["names"] = []
        if "name" in study:
            original_name = study["name"]
            study["names"].append(original_name)
            del study["name"]

        # Return the transformed dictionary.
        return study

    def create_report_based_upon_comment(self, comment: dict) -> dict:
        """
        Creates a report based upon the comment passed in.

        Note: Although this function will be passed a document from the `comment_set` collection, it actually modifies
              a _different_ (i.e. arbitrary) collection instead, and then returns the original `comment_set` document
              unchanged. I consider this to be a "shoehorned" usage of the migration framework, but a necessary
              usage since the migration framework doesn't provide a different way to do the same thing (yet).
        """
        report = {"body": f"Someone wrote {comment['text']}"}
        self.adapter.insert_document(collection_name="report_set", document=report)
        return comment  # returns the original `comment_set` document, unmodified



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset
import difflib
import uuid


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "X"
    _to_version = "PR19_and_PR70"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.create_collection(collection_name="instrument_set")
        self.fill_instrument_set()
      
        self.adapter.process_each_document(collection_name="omics_processing_set", pipeline=[self.transform_instrument_slot]) 

    def fill_instrument_set(self):
        r""""
        Fills the instrument_set collection from a YAML file of curated instrument instances
        
        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {"instrument_set": []}  # seeds the database
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.fill_instrument_set()
        >>> result = adapter.get_document_having_value_in_field('instrument_set', 'name', value='12T FT-ICR MS')
        >>> result['name']
        '12T FT-ICR MS'
        >>> result['vendor']
        'bruker'
        >>> result['model']
        'solarix_12T'
        >>> result['type']
        'nmdc:Instrument'

        """

        instruments = load_yaml_asset("migrator_from_X_to_PR19_and_PR70/instrument_set.yaml")
        for instrument in instruments:
            self.adapter.insert_document("instrument_set", instrument)

    def transform_instrument_slot(self, omics_doc: dict):
        r"""
        Changes the instrument_name slot to be instrument_used, and replaces the name string value with an identifier for 
        the instrument

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {'instrument_set': [{'id': 'nmdc:inst-456', 'name': '12T FT-ICR MS', 'model':'solarix_12T', 'vendor':'bruker', 'type':'nmdc:Instrument'}]} 
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-123', 'instrument_name': '12T_FTICR_B'})
        {'id': 'nmdc:omcp-123', 'instrument_used': ['nmdc:inst-456']}
        """

        if "instrument_name" in omics_doc:
            existing_instrument_name = omics_doc["instrument_name"]
            
            instruments = load_yaml_asset("migrator_from_X_to_PR19_and_PR70/instrument_set.yaml")
            allowed_instrument_names = [instrument["name"] for instrument in instruments]
            
            matches = difflib.get_close_matches(existing_instrument_name, allowed_instrument_names, n=1, cutoff=0.25)

            if matches:

                    # Match instrument_name with instrument set and get corresponding instrument id
                    match = matches[0]
                    self.logger.info(f"instrument_name in {omics_doc['id']} is {omics_doc['instrument_name']} and will be matched to instrument with name {match}")
                    
                    instrument = self.adapter.get_document_having_value_in_field(
                        collection_name="instrument_set", field_name="name", value=match
                    )

                    instrument_id = instrument.get("id")
                    omics_doc["instrument_used"] = [instrument_id]
                    del omics_doc["instrument_name"]
        
            else:
                self.logger.error(f"The instrument_name {omics_doc['id']} has a value of {omics_doc['instrument_name']}, but did not have any matches with the cutoff set at 0.25")

        return omics_doc


        


import unittest
from typing import Optional

from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter


class TestMongoAdapter(unittest.TestCase):
    r"""
    Tests targeting the `DictionaryAdapter` class.

    You can run these tests like this:
    $ poetry run python -m unittest -v nmdc_schema/migrators/adapters/test_dictionary_adapter.py

    References:
    - https://docs.python.org/3/library/unittest.html#basic-example
    - https://docs.python.org/3/library/unittest.html#unittest.TestCase.setUp
    """

    db: Optional[dict] = None

    def setUp(self) -> None:
        r"""
        Gets a reference to the database.

        Note: This function runs before each test starts.
        """
        self.db = dict()

    def tearDown(self) -> None:
        r"""
        Relinquishes the reference to the database.

        Note: This function runs after each test finishes.
        """
        self.db = None

    def test_create_collection(self):
        # Set up:
        collection_name = "my_collection"

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        adapter.create_collection(collection_name)

        # Validate result:
        assert len(self.db.keys()) == 1
        assert collection_name in self.db

    def test_rename_collection(self):
        # Set up:
        collection_name_original = "my_collection"
        collection_name_target = "your_collection"
        self.db[collection_name_original] = []
        assert len(self.db.keys()) == 1

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        adapter.rename_collection(collection_name_original, collection_name_target)

        # Validate result:
        assert len(self.db.keys()) == 1
        assert collection_name_target in self.db

    def test_delete_collection(self):
        # Set up:
        collection_name = "my_collection"
        self.db[collection_name] = []
        assert len(self.db.keys()) == 1

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        adapter.delete_collection(collection_name)

        # Validate result:
        assert len(self.db.keys()) == 0

    def test_insert_document(self):
        # Set up:
        collection_name = "my_collection"
        document = dict(id=123, foo="bar")
        self.db[collection_name] = []
        assert len(self.db[collection_name]) == 0

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        adapter.insert_document(collection_name, document)

        # Validate result:
        assert len(self.db[collection_name]) == 1
        document_actual = self.db[collection_name][0]
        for k, v in document.items():
            assert v == document_actual[k]

    def test_get_document_having_value_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        self.db[collection_name] = []
        self.db[collection_name].extend([document_1, document_2, document_3])
        assert len(self.db[collection_name]) == 3

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        document_actual = adapter.get_document_having_value_in_field(
            collection_name, "x", "b"
        )

        # Validate result:
        for k, v in document_2.items():
            assert v == document_actual[k]

    def test_get_document_having_one_of_values_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        self.db[collection_name] = []
        self.db[collection_name].extend([document_1, document_2, document_3])
        assert len(self.db[collection_name]) == 3

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        document_actual = adapter.get_document_having_one_of_values_in_field(
            collection_name, "x", ["n", "b", "m"]
        )

        # Validate result:
        for k, v in document_2.items():
            assert v == document_actual[k]

    def test_delete_documents_having_value_in_field(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(id=1, x="a")
        document_2 = dict(id=2, x="b")
        document_3 = dict(id=3, x="c")
        document_4 = dict(id=4, x="b")  # note: x="b" here also
        self.db[collection_name] = []
        self.db[collection_name].extend(
            [document_1, document_2, document_3, document_4]
        )
        assert len(self.db[collection_name]) == 4

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        num_documents_deleted = adapter.delete_documents_having_value_in_field(
            collection_name, "x", "b"
        )

        # Validate result:
        collection = self.db[collection_name]
        assert num_documents_deleted == 2
        assert len(self.db[collection_name]) == 2
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "a"]) == 1
        assert len([doc for doc in collection if doc["id"] == 2 and doc["x"] == "b"]) == 0
        assert len([doc for doc in collection if doc["id"] == 3 and doc["x"] == "c"]) == 1
        assert len([doc for doc in collection if doc["id"] == 4 and doc["x"] == "b"]) == 0

    def test_process_each_document(self):
        # Set up:
        collection_name = "my_collection"
        document_1 = dict(_id=1, id=1, x="a")
        document_2 = dict(_id=2, id=2, x="b")
        document_3 = dict(_id=3, id=3, x="c")
        self.db[collection_name] = []
        self.db[collection_name].extend(
            [document_1, document_2, document_3]
        )
        assert len(self.db[collection_name]) == 3

        def capitalize_x(doc: dict) -> dict:
            r"""Example pipeline stage that capitalizes the first letter of the `x` value."""
            doc["x"] = doc["x"].capitalize()
            return doc

        def append_z_to_x_value(doc: dict) -> dict:
            r"""Example pipeline stage that appends "z" to the `x` value."""
            doc["x"] = doc["x"] + "z"
            return doc

        # Invoke function-under-test:
        adapter = DictionaryAdapter(database=self.db)
        adapter.process_each_document(
            collection_name, [capitalize_x, append_z_to_x_value]
        )

        # Validate result:
        collection = self.db[collection_name]
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "a"]) == 0  # pre-pipeline
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "b"]) == 0
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "c"]) == 0
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "A"]) == 0  # mid-pipeline
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "B"]) == 0
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "C"]) == 0
        assert len([doc for doc in collection if doc["id"] == 1 and doc["x"] == "Az"]) == 1  # post-pipeline
        assert len([doc for doc in collection if doc["id"] == 2 and doc["x"] == "Bz"]) == 1
        assert len([doc for doc in collection if doc["id"] == 3 and doc["x"] == "Cz"]) == 1

    def test_callbacks(self):
        # Set up:
        collection_name = "my_collection"
        new_collection_name = "my_new_collection"
        event_log: list[str] = []
        adapter = DictionaryAdapter(
            database=self.db,
            on_collection_created=lambda name: event_log.append(f"Created {name}"),
            on_collection_renamed=lambda old_name, name: event_log.append(f"Renamed {old_name} to {name}"),
            on_collection_deleted=lambda name: event_log.append(f"Deleted {name}"),
        )
        assert len(event_log) == 0

        # Trigger callback and validate latest message in event log:
        adapter.create_collection(collection_name)
        assert len(event_log) == 1
        assert event_log[0] == f"Created {collection_name}"

        # Trigger callback and validate latest message in event log:
        adapter.rename_collection(collection_name, new_collection_name)
        assert len(event_log) == 2
        assert event_log[1] == f"Renamed {collection_name} to {new_collection_name}"

        # Trigger callback and validate latest message in event log:
        adapter.delete_collection(new_collection_name)
        assert len(event_log) == 3
        assert event_log[2] == f"Deleted {new_collection_name}"


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.nmdc_data import get_nmdc_schema_definition
from linkml_runtime import SchemaView


def create_schema_view() -> SchemaView:
    """
    Returns a LinkML SchemaView instance that can be used to traverse the schema.
    
    >>> isinstance(create_schema_view(), SchemaView)
    True
    """
    schema_definition = get_nmdc_schema_definition()
    schema_view = SchemaView(schema_definition)
    return schema_view


class Migrator(MigratorBase):
    """Migrates data from schema X to PR10"""

    _from_version = "X"
    _to_version = "PR10"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        # Get a dictionary of slots and the class uris of their range if they have inlined classes as their range.
        view = create_schema_view()

        slots_with_inlined_classes = {}
        classes_with_inlined_classes = ["Biosample", "Study", "Extraction", "MetabolomicsAnalysis" , "MetaproteomicsAnalysis", "MagsAnalysis"]
        for nmdc_class in classes_with_inlined_classes:
            induced_slots = view.class_induced_slots(nmdc_class)
            for slot_def in induced_slots:
                slot_name = slot_def.name
                slot_range = slot_def.range
                element = view.get_element(slot_range)
                metatype = type(element).class_name
                if metatype == 'class_definition':
                    class_identifying_slot = view.get_identifier_slot(slot_range)
                    if not class_identifying_slot:
                        # need to hardcode because schema prefix is not NMDC
                        if slot_name == "has_credit_associations":
                            class_uri = "prov:Association"
                            slots_with_inlined_classes[slot_name] = class_uri
                        else:
                            class_uri = f"{view.schema.default_prefix}:{element.name}"
                            slots_with_inlined_classes[slot_name] = class_uri

        # Populate the "collection-to-transformers" map for this specific migration.
        agenda = dict(
            biosample_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Biosample", slots_with_inlined_classes)],
            data_object_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:DataObject")],
            functional_annotation_agg=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:FunctionalAnnotationAggMember")],
            study_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Study", slots_with_inlined_classes)],
            extraction_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Extraction", slots_with_inlined_classes)],
            field_research_site_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:FieldResearchSite")],
            library_preparation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:LibraryPreparation")],
            mags_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MagsAnalysis", slots_with_inlined_classes)],
            metabolomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetabolomicsAnalysis", slots_with_inlined_classes)],
            metagenome_annotation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAnnotation")],
            metagenome_assembly_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAssembly")],
            metagenome_sequencing_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeSequencing")],
            metaproteomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetaproteomicsAnalysis", slots_with_inlined_classes)],
            metatranscriptome_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeAnalysis")],
            nom_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:NomAnalysis")],
            data_generation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:DataGeneration")],
            pooling_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Pooling")],
            processed_sample_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ProcessedSample")],
            read_based_taxonomy_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadBasedTaxonomyAnalysis")],
            read_qc_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadQcAnalysis")],
        )

        for collection_name, pipeline in agenda.items():
            self.adapter.process_each_document(collection_name=collection_name, pipeline=pipeline)

    def add_type_to_inlined_classes(self, document: dict, slot: str, uri: str):
        r"""
        Adds a type slot to each inlined instance of an NMDC class in the biosample_set, study_set, extraction_set, 
        mags_set, metabolomics_analysis_set, and metaproteomics_analysis_set. 
        """
        if document.get(slot):
            # If slot is a list, iterate over each item in the list (e.g. chem_administration and has_credit_associations)
            if isinstance(document[slot], list):
                for item in document[slot]:
                    # will exclude non-inlined class slots that use an id
                    if not isinstance(item, str):
                        item["type"] = uri
                        if item.get("term"):
                            item["term"]["type"] = "nmdc:OntologyClass"
                        if item.get("applies_to_person"):
                            item["applies_to_person"]["type"] = "nmdc:PersonValue"
            else:
                # If slot is not multivalued
                document[slot]["type"] = uri
                if document[slot].get("term"):
                    document[slot]["term"]["type"] = "nmdc:OntologyClass"
                    

    def add_type_slot_with_class_uri(self, document: dict, class_uri: str, inlined_slots: Optional[dict] = None):
        r"""
            Adds a type slot to each collection with the appropriate class uri as the value. E.g. type: nmdc:Biosample. If a type
            slot exists, it will overwrite to the types listed below. If the optional dictionary argument inlined_slots is given, then
            this function will add the type lost to any slots that have inlined classes as their range.
        
            >>> m = Migrator()
            >>> m.add_type_slot_with_class_uri({'id': 123, 'collection_date': {'has_raw_value': '2017-05-09'}}, 'nmdc:Biosample', {'collection_date': 'nmdc:TimestampValue'}) 
            {'id': 123, 'collection_date': {'has_raw_value': '2017-05-09', 'type': 'nmdc:TimestampValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'type': 'nmdc:DataGeneration'}, 'nmdc:DataGeneration')
            {'id': 567, 'type': 'nmdc:DataGeneration'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234'}}}, 'nmdc:Biosample', {'env_broad_scale': 'nmdc:ControlledIdentifiedTermValue'})
            {'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234', 'type': 'nmdc:OntologyClass'}, 'type': 'nmdc:ControlledIdentifiedTermValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 456}, 'nmdc:NomAnalysis')
            {'id': 456, 'type': 'nmdc:NomAnalysis'}
        """
        
        # Adds the type slot with the correct class_uri as a value to each collection instance
        document["type"] = class_uri
                
        # Add the type slot to any inlined classes in the Biosample, Study, or Extraction collections
        if inlined_slots:
            for slot, uri in inlined_slots.items():
                self.add_type_to_inlined_classes(document, slot, uri)
                
        return document

    


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR2 and PR24, namely renames all renamed collections.
    """
    _from_version = "X"
    _to_version = "PR2_and_PR24"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        old_to_new_names = {
            "omics_processing_set": "data_generation_set",
            "mags_activity_set": "mags_set",
            "metabolomics_analysis_activity_set": "metabolomics_analysis_set",
            "metagenome_annotation_activity_set": "metagenome_annotation_set",
            "metagenome_sequencing_activity_set": "metagenome_sequencing_set",
            "metaproteomics_analysis_activity_set": "metaproteomics_analysis_set",
            "metatranscriptome_activity_set": "metatranscriptome_analysis_set",
            "nom_analysis_activity_set": "nom_analysis_set",
            "read_based_taxonomy_analysis_activity_set": "read_based_taxonomy_analysis_set",
            "read_qc_analysis_activity_set": "read_qc_analysis_set",
            "activity_set": "workflow_execution_set"
            }

        for current_collection_name, new_collection_name in old_to_new_names.items():
            try:
                self.adapter.rename_collection(current_name=current_collection_name, new_name=new_collection_name)
            except KeyError:
                print(f"Error: Collection '{current_collection_name}' not found in the adapter.")




from nmdc_schema.migrators.migrator_base import MigratorBase
import difflib


class Migrator(MigratorBase):
    """Migrates data from schema X to PR23"""

    _from_version = "X"
    _to_version = "PR23"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        # Populate the "collection-to-transformers" map for this specific migration.
        agenda = dict(
            metagenome_assembly_set=[self.standardize_execution_resource], 
            metagenome_annotation_activity_set=[self.standardize_execution_resource],
            metatranscriptome_activity_set=[self.standardize_execution_resource], 
            mags_activity_set=[self.standardize_execution_resource],
            metagenome_sequencing_activity_set=[self.standardize_execution_resource], 
            read_qc_analysis_activity_set=[self.standardize_execution_resource],
            read_based_taxonomy_analysis_activity_set=[self.standardize_execution_resource], 
            metabolomics_analysis_activity_set=[self.standardize_execution_resource],
            metaproteomics_analysis_activity_set=[self.standardize_execution_resource], 
            nom_analysis_activity_set=[self.standardize_execution_resource]
        )

        for collection_name, pipeline in agenda.items():
            self.adapter.process_each_document(collection_name=collection_name, pipeline=pipeline)

    def standardize_execution_resource(self, workflow_execution: dict):
        r"""
        Transforms the values of the execution_resource slot for any subclasses of WorkflowExecutionActivity into the
        allowable permissible values of the ExecutionResourceEnum.
        
        >>> m = Migrator()
        >>> m.standardize_execution_resource({'id': 123, 'execution_resource': 'LANL B-div'})
        {'id': 123, 'execution_resource': 'LANL-B-div'}
        >>> m.standardize_execution_resource({'id': 234, 'execution_resource': 'NERSC - Cori'})
        {'id': 234, 'execution_resource': 'NERSC-Cori'}
        """

        workflow_id = workflow_execution["id"]

        # allowable permissible values
        allowed_execution_resources = ['NERSC-Cori', 'NERSC-Perlmutter', 'EMSL', 'EMSL-RZR', 'JGI', 'LANL-B-div']
        
        if 'execution_resource' in workflow_execution:
            
            workflow_resource_value = workflow_execution['execution_resource']

            # Get the permissible value, if any, that is most similar to the original value
            # and has a similarity score of at least 0.8 with respect to the original value.
            # Reference: https://docs.python.org/3/library/difflib.html#difflib.get_close_matches
            matches = difflib.get_close_matches(workflow_resource_value, allowed_execution_resources, n=1, cutoff=0.8)
            
            if matches:

                if workflow_resource_value != matches[0]:
                    # change value of the workflow_execution resource to the proper permissible value if the match is not 100%
                    self.logger.info(f"Changing the workflow {workflow_id} execution_resource value from {workflow_resource_value} to {matches[0]}")
                    workflow_execution['execution_resource'] = matches[0]

            else:
                self.logger.error(f"The workflow {workflow_id} has an execution_resource value of {workflow_resource_value}, but did not have any matches with the cutoff set at 0.8")

        return workflow_execution

    


from abc import ABC, abstractmethod
from typing import Dict, List
from logging import getLogger
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase


class MigratorBase(ABC):
    """Base class containing properties and methods related to migrating data between two schema versions."""

    # The schema version from which this class migrates data.
    #
    # Note: This string is empty here. It will be populated within the migration-specific classes.
    #
    _from_version: str = ""

    # The schema version to which this class migrates data.
    #
    # Note: This string is empty here. It will be populated within the migration-specific classes.
    #
    _to_version: str = ""

    def __init__(self, adapter: AdapterBase = None, logger=None):
        # Store a reference to the specified adapter. Migrator methods can use it to manipulate the database.
        self.adapter = adapter

        # If a logger was specified, use it; otherwise, initialize one and use that.
        self.logger = getLogger(__name__) if logger is None else logger

        if self.adapter is None:
            self.logger.warning("No adapter was specified. Migration capability will be limited.")

    @abstractmethod
    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        pass

    @classmethod
    def get_origin_version(cls) -> str:
        """Returns the schema version this class migrates data from."""
        return cls._from_version

    @classmethod
    def get_destination_version(cls) -> str:
        """Returns the schema version this class migrates data to."""
        return cls._to_version




from linkml_runtime import SchemaView

schema_file = "../src/schema/nmdc.yaml"

schema_view = SchemaView(schema_file)

print(schema_view.schema.name)

schema_classes = schema_view.all_classes()

for ck, cv in schema_classes.items():
    cis = schema_view.class_induced_slots(ck)
    for induced_slot in cis:

        if "structured_pattern" in induced_slot and induced_slot["structured_pattern"]:
            sp_syntax = induced_slot["structured_pattern"]["syntax"]
            print(f"{ck} {induced_slot.name}: {sp_syntax}")



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase
from difflib import SequenceMatcher

class Migrator(MigratorBase):
    r"""
    Migrates data from X to PR31, removes used slot from WorkflowExecution subclasses and checks that the 
    value in the used slot on the WorkflowExecution classes matches the value on the DataGeneration 
    instances in the instrument_name slot.
    """

    _from_version = "X"
    _to_version = "PR31"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        
        workflow_execution_collection_names = [
            "mags_activity_set",
            "metabolomics_analysis_activity_set",
            "metagenome_annotation_activity_set",
            "metagenome_assembly_set",
            "metagenome_sequencing_activity_set",
            "metatranscriptome_activity_set",
            "nom_analysis_activity_set",
            "omics_processing_set",
            "read_based_taxonomy_analysis_activity_set",
            "read_qc_analysis_activity_set"
            "metaproteomics_analysis_activity_set"   
        ]
        
        for collection_name in workflow_execution_collection_names:
            self.adapter.process_each_document(
                collection_name=collection_name,
                pipeline=[self.remove_used_slot],
            )
    
    def preprocess_string(self, s):
        r"""
        Normalizes strings prior to using SequenceMatcher. Removes white spaces, hyphens, and 
        underscores from a string so difflib's SequenceMatcher can find the longest contiguous 
        matching subsequence between two sequences and these characters will not interfere.
        >>> m = Migrator()
        >>> m.preprocess_string('a  b_-_c -de:f g')
        'abcde:fg'
        """

        return s.replace(" ", "").replace("_","").replace("-","")

    def remove_used_slot(self, doc: dict) -> dict:
        r"""
        Removes the `used` slot from `WorkflowExecution` subclasses if the value matches the 
        instrument_name slot from the corresponding `OmicsProcessing` document by the longest
        common sequence.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {'omics_processing_set':[{'id':'nmdc:omcp-123', 'instrument_name':'nmdc:wfc-456'}]}  # in this example, our data store is a Python dictionary
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.remove_used_slot({'id': 'nmdc:metab-123', 'used': 'nmdc:wfc-456', 'was_informed_by': 'nmdc:omcp-123'})
        {'id': 'nmdc:metab-123', 'was_informed_by': 'nmdc:omcp-123'}
        """
        
        if "used" in doc:
            omics_processing_doc = self.adapter.get_document_having_value_in_field(
                collection_name="omics_processing_set", field_name="id", value=doc["was_informed_by"]
            )
            
            # Preprocess instrument strings to ignore hyphens, underscores, and blank spaces
            processed_workflow_instrument_string = self.preprocess_string(doc["used"])
            processed_omics_doc_instrument_string = self.preprocess_string(omics_processing_doc["instrument_name"])
            
            similarity_ratio = SequenceMatcher(None, processed_workflow_instrument_string, processed_omics_doc_instrument_string).ratio()
            threshold = 0.8
            if similarity_ratio >= threshold:
                if similarity_ratio < 1.0:
                    self.logger.info(f"Workflow with id {doc['id']} has instrument: {doc['used']} matches OmicsProcessing doc instrument: {omics_processing_doc['instrument_name']} well enough")
                doc.pop("used")
            else:
                self.logger.error(f"Workflow doc {doc['id']} with instrument: {doc['used']} does not match {omics_processing_doc['instrument_name']}")
        
        return doc

    



from linkml_runtime.dumpers import yaml_dumper

from nmdc_schema.nmdc_schema_accepting_legacy_ids import Biosample, ControlledIdentifiedTermValue, OntologyClass


def do_test():
    bs = Biosample(id='xxx',
                   type='nmdc:Biosample',
                   associated_studies='gold:Gs0114663',
                   env_medium=ControlledIdentifiedTermValue(
                       type='nmdc:ControlledIdentifiedTermValue',
                       term=OntologyClass(
                           type='nmdc:OntologyClass',
                           id='ENVO:00010483',
                           name='freshwater environment',
                       ),
                   ),
                   env_broad_scale=ControlledIdentifiedTermValue(
                       type='nmdc:ControlledIdentifiedTermValue',
                       term=OntologyClass(
                           type='nmdc:OntologyClass',
                           id='ENVO:00010483',
                           name='freshwater environment',
                       ),
                   ),
                   env_local_scale=ControlledIdentifiedTermValue(
                       type='nmdc:ControlledIdentifiedTermValue',
                       term=OntologyClass(
                           type='nmdc:OntologyClass',
                           id='ENVO:00010483',
                           name='freshwater environment',
                       ),
                   ),
                   )

    print(yaml_dumper.dumps(bs))



from typing import Dict, List

from nmdc_schema.migrators.helpers import load_yaml_asset
from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.adapters.adapter_base import AdapterBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR9.

    Namely, creates the `workflow_chain_set` collection, moves `was_informed_by` onto the `WorkflowChain` instances,
    and changes the `part_of` slots on instances of subclasses of `WorkflowExecutionActivity` to the id of the
    corresponding `WorkflowChain`.
    """

    _from_version = "X"
    _to_version = "PR9"

    def __init__(self, adapter: AdapterBase = None, logger=None):
        r"""
        Initializes some empty dictionaries and loads minted `WorkflowChain` ids from a YAML file into a list.
        """

        # Invoke the `__init__` method of the "parent" class.
        #
        # Note: In Python, when a "child" (i.e. inheriting) class doesn't have its own `__init__` method,
        #       it inherits the "parent" class's `__init__` method. As a result, that inherited `__init__`
        #       method is the one that Python automatically invokes when the "child" class is instantiated.
        #
        #       In contrast, _this_ "child" class _does_ have its own `__init__` method (this comment is in it),
        #       effectively opting the "child" class out of the aforementioned inheritance and automatic invocation
        #       of the "parent" class's `__init__` method. So, here, we invoke that method "manually."
        #
        super().__init__(adapter=adapter, logger=logger)

        self.omics_proc_id_to_workflow_chain_id: Dict[str, str] = {}
        self.omics_proc_id_to_analyte_category: Dict[str, str] = {}
        self.workflow_chain_ids: List[str] = load_yaml_asset("migrator_from_X_to_PR9/workflow_chain_ids.yaml")

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.
        """

        workflow_execution_activity_collection_names = [
            "metagenome_sequencing_activity_set",
            "read_qc_analysis_activity_set",
            "metagenome_assembly_set",
            "read_based_taxonomy_analysis_activity_set",
            "metagenome_annotation_activity_set",
            "mags_activity_set",
            "metabolomics_analysis_activity_set",
            "nom_analysis_activity_set",
            "metatranscriptome_activity_set",
            "metaproteomics_analysis_activity_set",
        ]

        # Populate the dictionary that maps `OmicsProcessing` IDs to `WorkflowChain` IDs (using the `was_informed_by`
        # field of the `WorkflowExecutionActivity` document); then, store those `WorkflowChain` IDs in the `part_of`
        # field of `WorkflowExecutionActivity` documents.
        for collection_name in workflow_execution_activity_collection_names:
            self.adapter.process_each_document(
                collection_name=collection_name,
                pipeline=[self.generate_mapping_from_omics_proc_id_to_workflow_chain_id, self.update_part_of_field],
            )

        self.logger.info(f"{len(self.omics_proc_id_to_workflow_chain_id)=}")

        # Populate the dictionary that maps `OmicsProcessing` IDs to their analyte categories.
        self.adapter.process_each_document(
            collection_name="omics_processing_set",
            pipeline=[self.generate_mapping_from_omics_proc_id_to_analyte_category],
        )

        # Create and populate the `workflow_chain_set` collection.
        self.adapter.create_collection("workflow_chain_set")
        self.populate_workflow_chain()

        # Remove the `was_informed_by` field from the `WorkflowExecutionActivity` documents
        # if they match in their corresponding `WorkflowChain` records.
        for collection_name in workflow_execution_activity_collection_names:
            self.adapter.process_each_document(
                collection_name=collection_name,
                pipeline=[self.remove_was_informed_by_field],
            )

    def generate_mapping_from_omics_proc_id_to_workflow_chain_id(self, workflow_execution_activity: dict):
        r"""
        Gets the `OmicsProcessing` ID from the `was_informed_by` field of the `WorkflowExecutionActivity` document
        and — if not already done — adds it to the dictionary that maps `OmicsProcessing` IDs to `WorkflowChain` IDs.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> m = Migrator(adapter=DictionaryAdapter(database={}))
        >>> m.generate_mapping_from_omics_proc_id_to_workflow_chain_id({'id': 123, 'was_informed_by': 'nmdc:omcp-123'})
        {'id': 123, 'was_informed_by': 'nmdc:omcp-123'}
        >>> 'nmdc:omcp-123' in m.omics_proc_id_to_workflow_chain_id
        True
        """

        omics_processing_id = workflow_execution_activity["was_informed_by"]
        if omics_processing_id not in self.omics_proc_id_to_workflow_chain_id.keys():

            # Claim a minted `WorkflowChain` ID — if we have any unclaimed ones — for this `OmicsProcessing` document.
            if len(self.workflow_chain_ids) > 0:
                workflow_chain_id = self.workflow_chain_ids.pop(0)  # removes and returns the first item in the list
                self.omics_proc_id_to_workflow_chain_id[omics_processing_id] = workflow_chain_id
            else:
                self.logger.error(f"No WorkflowChain ID available for OmicsProcessing: {omics_processing_id}")

        return workflow_execution_activity

    def update_part_of_field(self, workflow_execution_activity: dict):
        r"""
        Populates (replacing the existing value, if any) the `part_of` field of the `WorkflowExecutionActivity`
        document, with a list consisting of the `OmicsProcessing` ID from its own `was_informed_by` field.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> m = Migrator(adapter=DictionaryAdapter(database={}))
        >>> m.omics_proc_id_to_workflow_chain_id = {'nmdc:omcp-123': 'nmdc:wfc-456'}
        >>> result1 = m.update_part_of_field({'id': 456, 'was_informed_by': 'nmdc:omcp-123'})
        >>> result1
        {'id': 456, 'was_informed_by': 'nmdc:omcp-123', 'part_of': ['nmdc:wfc-456']}
        >>> result2 = m.update_part_of_field({'id': 456, 'was_informed_by': 'nmdc:omcp-123', 'part_of': ['nmdc:789']})
        >>> result2
        {'id': 456, 'was_informed_by': 'nmdc:omcp-123', 'part_of': ['nmdc:wfc-456']}
        """

        omics_processing_id = workflow_execution_activity["was_informed_by"]
        workflow_chain_id = self.omics_proc_id_to_workflow_chain_id.get(omics_processing_id)
        workflow_execution_activity["part_of"] = [workflow_chain_id]

        return workflow_execution_activity

    def generate_mapping_from_omics_proc_id_to_analyte_category(self, omics_doc: dict):
        r"""
        Generates a mapping from `OmicsProcessing` IDs to analyte categories.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> m = Migrator(adapter=DictionaryAdapter(database={}))
        >>> omics_doc = {'id': 'nmdc:omcp-123', 'analyte_category': 'metagenome'}
        >>> result = m.generate_mapping_from_omics_proc_id_to_analyte_category(omics_doc)
        >>> result
        {'id': 'nmdc:omcp-123', 'analyte_category': 'metagenome'}
        >>> m.omics_proc_id_to_analyte_category
        {'nmdc:omcp-123': 'metagenome'}
        """

        omics_doc_id = omics_doc["id"]
        analyte_category = omics_doc["analyte_category"]

        self.omics_proc_id_to_analyte_category[omics_doc_id] = analyte_category

        return omics_doc

    def populate_workflow_chain(self) -> None:
        r"""
        Populates the `workflow_chain_set` collection based upon the mappings defined in this class's initializer.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {"workflow_chain_set": []}  # seeds the database
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.omics_proc_id_to_workflow_chain_id = {'nmdc:omcp-123': 'nmdc:wfc-456'}
        >>> m.omics_proc_id_to_analyte_category = {'nmdc:omcp-123': 'metagenome'}
        >>> m.populate_workflow_chain()
        >>> adapter.get_document_having_value_in_field('workflow_chain_set', 'id', value='nmdc:wfc-456')
        {'id': 'nmdc:wfc-456', 'was_informed_by': 'nmdc:omcp-123', 'analyte_category': 'metagenome', 'type': 'nmdc:WorkflowChain', 'name': 'Workflow Chain for metagenome analysis of nmdc:omcp-123'}
        """

        for omics_id, workflow_chain_id in self.omics_proc_id_to_workflow_chain_id.items():
            workflow_chain_doc = {}

            analyte_category = None
            if omics_id in self.omics_proc_id_to_analyte_category:
                analyte_category = self.omics_proc_id_to_analyte_category.get(omics_id)

            workflow_chain_doc["id"] = workflow_chain_id
            workflow_chain_doc["was_informed_by"] = omics_id
            workflow_chain_doc["analyte_category"] = analyte_category
            workflow_chain_doc["type"] = "nmdc:WorkflowChain"
            workflow_chain_doc["name"] = (
                f"Workflow Chain for {analyte_category} analysis of {omics_id}"
            )

            self.adapter.insert_document("workflow_chain_set", workflow_chain_doc)

    def remove_was_informed_by_field(self, workflow_doc: dict):
        r"""
        Removes the was_informed_by slot from the workflow sets if its value matches its corresponding 
        workflow_chain_set doc's was_informed_by value.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {'workflow_chain_set': [{'id': 'nmdc:wfc-456', 'was_informed_by': 'nmdc:omcp-123'}]}  # in this example, our data store is a Python dictionary
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.remove_was_informed_by_field({'id': 'nmdc:metab-123', 'part_of': ['nmdc:wfc-456'], 'was_informed_by': 'nmdc:omcp-123'})
        {'id': 'nmdc:metab-123', 'part_of': ['nmdc:wfc-456']}
        """

        omics_id = workflow_doc["was_informed_by"]
        workflow_chain_ids = workflow_doc["part_of"]
        
        for wfc_id in workflow_chain_ids:

            workflow_chain_doc = self.adapter.get_document_having_value_in_field(
                collection_name="workflow_chain_set", field_name="id", value=wfc_id
            )
            self.logger.info(f"{workflow_chain_doc['id']=}")

            if workflow_chain_doc["was_informed_by"] == omics_id:
                del workflow_doc["was_informed_by"]
            else:
                self.logger.error(
                    f"WorkflowExecution doc with id {workflow_doc['id']} was_informed_by slot does not match"
                    f"its workflow chain doc with id {workflow_chain_doc['id']} was_informed_by slot"
                )

        return workflow_doc



import re
from pathlib import Path
from functools import wraps

import click

# TODO: Determine directory path dynamically, instead of hard-coding it.
MIGRATOR_DIRECTORY = "nmdc_schema/migrators"


def click_option_validator(function_to_decorate):
    r"""
    This decorator wraps a function in another function, which strips away the
    first two arguments and then passes the third argument to the wrapped function.
    """

    # Note: The `@wraps(fn)` decorator makes the wrapper function "look like" the wrapped
    #       function; in that the wrapper will adopt the latter's name and docstring.
    #       Reference: https://docs.python.org/3/library/functools.html#functools.wraps
    @wraps(function_to_decorate)
    def wrapper(click_context, param, param_value):
        return function_to_decorate(param_value)

    return wrapper


@click_option_validator
def validate_version_identifier(version_identifier: str) -> str:
    r"""
    Validates a version identifier, raising an exception if it's invalid.

    Reference: https://click.palletsprojects.com/en/8.1.x/options/#callbacks-for-validation

    >>> validate_version_identifier(None, None, "1.2.3")
    '1.2.3'
    >>> validate_version_identifier(None, None, "1.2.")  # doesn't end with a number
    Traceback (most recent call last):
    ...
    click.exceptions.BadParameter: Invalid format.
    >>> validate_version_identifier(None, None, "1.2")
    '1.2'
    >>> validate_version_identifier(None, None, "123.456.789")
    '123.456.789'
    >>> validate_version_identifier(None, None, "123.456.789.1")  # 4 parts
    Traceback (most recent call last):
    ...
    click.exceptions.BadParameter: Invalid format.
    >>> validate_version_identifier(None, None, "123")
    '123'
    """
    if re.fullmatch(r"^(\d+)(\.\d+){0,2}$", version_identifier) is None:
        raise click.BadParameter("Invalid format.")

    return version_identifier


@click.command()
@click.option(
    "--from-version",
    "--from",
    type=str,
    required=True,
    help='Schema version from which the migrator will migrate data (e.g. "1.0").',
    prompt="From schema version",
    callback=validate_version_identifier,
)
@click.option(
    "--to-version",
    "--to",
    type=str,
    required=True,
    prompt="To schema version",
    help='Schema version to which the migrator will migrate data (e.g. "1.1").',
    callback=validate_version_identifier,
)
def create_migrator(from_version: str, to_version: str) -> None:
    """
    Create a migrator class based upon the specified version identifiers.
    """

    # Generate the migrator module name.
    from_version_snake = from_version.replace(".", "_")
    to_version_snake = to_version.replace(".", "_")
    file_name = f"migrator_from_{from_version_snake}_to_{to_version_snake}.py"

    # Generate the migrator file contents.
    # TODO: Move this to a dedicated template file to facilitate maintenance.
    file_contents = f'''
from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "{from_version}"
    _to_version = "{to_version}"    
    
    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""
        pass

'''

    # Create and populate the file (unless it already exists).
    file_path = Path(MIGRATOR_DIRECTORY) / Path(file_name)

    try:
        # Note: The "x" stands for "exclusive creation" mode, a mode in
        #       which an exception is thrown if the file already exists.
        with file_path.open("x") as file:
            file.write(file_contents)
        click.echo(f"Created '{file_name}' successfully.")
    except FileExistsError:
        click.echo(f"Error: File '{file_name}' already exists: {file_path}", err=True)
    except Exception as e:
        click.echo(f"Error: {e}", err=True)

    return None


if __name__ == "__main__":
    create_migrator()



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "10.0.0"  
    _to_version = "10.1.2"  

    def upgrade(self) -> None:
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        # Update each document in the `data_object_set` collection so that it uses a valid `FileTypeEnum` value.
        #
        # Note: The schema changed such that the `FileTypeEnum` value, "Metagenome Bins Compression File", is no longer
        #       valid. It was replaced with the value, "Metagenome HQMQ Bins Compression File". This migrator
        #       transforms all occurrences of the former value into the latter value, so that data that was valid with
        #       respect to the old schema is valid with respect to the new schema. According to the schema docs,
        #       the `FileTypeEnum` enum is used only by the `data_object_type` slot of the `DataObject` class.
        #
        self.adapter.process_each_document(
            "data_object_set", [self.update_data_object_type]
        )

    def update_data_object_type(self, data_object: dict) -> dict:
        r"""
        Updates a data object's `data_object_type` value to "Metagenome HQMQ Bins Compression File",
        if its value was originally "Metagenome Bins Compression File".

        >>> m = Migrator()
        >>> m.update_data_object_type({'id': 1})  # test: field not present
        {'id': 1}
        >>> m.update_data_object_type({'id': 1, 'data_object_type': 'unrelated'})  # test: value not relevant
        {'id': 1, 'data_object_type': 'unrelated'}
        >>> m.update_data_object_type({'id': 1, 'data_object_type': 'Metagenome HQMQ Bins Compression File'})
        {'id': 1, 'data_object_type': 'Metagenome HQMQ Bins Compression File'}
        >>> m.update_data_object_type({'id': 1, 'data_object_type': 'Metagenome Bins Compression File'})
        {'id': 1, 'data_object_type': 'Metagenome HQMQ Bins Compression File'}
        """
        self.logger.info(f"Processing DataObject: {data_object['id']}")

        invalid_value = "Metagenome Bins Compression File"
        valid_value = "Metagenome HQMQ Bins Compression File"

        # If this data object's `data_object_type` consists of the invalid value, replace it with the valid value.
        if "data_object_type" in data_object:
            if data_object["data_object_type"] == invalid_value:
                data_object["data_object_type"] = valid_value

        return data_object



import pprint

from linkml_runtime import SchemaView

schema_file = '../nmdc_schema/nmdc_schema_merged.yaml'

schema_view = SchemaView(schema_file)

schema_elements = schema_view.all_elements()

constraints_dict = {}
for ek, ev in schema_elements.items():
    et = type(ev).__name__
    current_key = f"{et} {ek}"
    if '_at_time' in ek:
        continue
    if 'id_prefixes' in ev and ev['id_prefixes']:
        current_id_prefixes = ev['id_prefixes']
        current_id_prefixes.sort()
        if current_key in constraints_dict:
            constraints_dict[current_key]['id_prefixes'] = current_id_prefixes
        else:
            constraints_dict[current_key] = {'id_prefixes': current_id_prefixes}
    if 'pattern' in ev and ev['pattern']:
        if current_key in constraints_dict:
            constraints_dict[current_key]['pattern'] = ev['pattern']
        else:
            constraints_dict[current_key] = {'pattern': ev['pattern']}
    if 'slot_usage' in ev and ev['slot_usage']:
        for uk, uv in ev['slot_usage'].items():
            current_key = f"{uk} used in {ek}"
            if 'id_prefixes' in uv and uv['id_prefixes']:
                current_id_prefixes = uv['id_prefixes']
                current_id_prefixes.sort()
                if current_key in constraints_dict:
                    constraints_dict[current_key]['id_prefixes'] = current_id_prefixes
                else:
                    constraints_dict[current_key] = {'id_prefixes': current_id_prefixes}
            if 'pattern' in uv and uv['pattern']:
                if current_key in constraints_dict:
                    constraints_dict[current_key]['pattern'] = uv['pattern']
                else:
                    constraints_dict[current_key] = {'pattern': uv['pattern']}

pprint.pprint(constraints_dict)



from nmdc_schema.migrators.migrator_base import MigratorBase
import re


class Migrator(MigratorBase):
    """Migrates data from schema X to PR53"""

    _from_version = "X"
    _to_version = "PR53"
    
    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document(collection_name="omics_processing_set",
                                           pipeline=[self.move_part_of_to_associated_studies])
        self.adapter.process_each_document(collection_name="biosample_set",
                                           pipeline=[self.move_part_of_to_associated_studies])

    def move_part_of_to_associated_studies(self, doc: dict):
        r"""
        Moves `part_of` values that are studies, into a new slot called `associated_studies`
        
        >>> m = Migrator()
        >>> m.move_part_of_to_associated_studies({'id': 123, 'part_of': ['gold:Gs0114663', 'nmdc:sty-55-xxx']})
        {'id': 123, 'associated_studies': ['gold:Gs0114663', 'nmdc:sty-55-xxx']}
        >>> m.move_part_of_to_associated_studies({'id': 123, 'part_of': []})  # `part_of` list is empty
        {'id': 123}
        >>> m.move_part_of_to_associated_studies({'id': 123})  # lacks `part_of` key
        {'id': 123}
        """

        # Check whether the document has a "part_of" field.
        if "part_of" in doc:

            # If the "part_of" field contains any study identifiers, add an "associated_studies" list field
            # to the document and store each study identifier in that list.
            study_ids = doc["part_of"]
            if len(study_ids) > 0:
                doc["associated_studies"] = []
                for study_id in study_ids:
                    doc["associated_studies"].append(study_id)

            # Delete the "part_of" field.
            del doc["part_of"]

        return doc



from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset


class Migrator(MigratorBase):
    """Migrates data from schema X to PR21"""

    _from_version = "X"
    _to_version = "PR21"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document(collection_name="study_set",
                                           pipeline=[self.move_relevant_protocols_to_protocol_link])

    def move_relevant_protocols_to_protocol_link(self, study: dict):
        r"""
        Moves any urls that exist in the relevant_protocols slot to a dictionary format (e.g. Protocol class is the range) 
        in the protocol_link slot. It then removes the relevant_protocols slot from the study.
        
        >>> m = Migrator()
        >>> m.move_relevant_protocols_to_protocol_link({'id': 123, 'relevant_protocols': ['www.ex1.org', 'www.ex2.gov']})
        {'id': 123, 'protocol_link': [{'url': 'www.ex1.org', 'type': 'nmdc:Protocol'}, {'url': 'www.ex2.gov', 'type': 'nmdc:Protocol'}]}
        >>> m.move_relevant_protocols_to_protocol_link({'id': 123, 'relevant_protocols': ['www.ex3.org']})
        {'id': 123, 'protocol_link': [{'url': 'www.ex3.org', 'type': 'nmdc:Protocol'}]}
        """

        if "relevant_protocols" in study:
            id = study["id"]
            study["protocol_link"] = []
            for protocol in study["relevant_protocols"]:
                new_protocol_struct = {
                    "url": protocol,
                    "type": "nmdc:Protocol"
                }
                study["protocol_link"].append(new_protocol_struct)
            self.logger.info(f"Moving {protocol} for study{id} from relevant_protocols slot to protocol_link")
        
            study.pop("relevant_protocols")  

        
        return study

    